%% Filtering joint angles
%% Author: Nayan Bhatt, IIT-Madras. Date: 28th January 2017.
%% Code Description: Kinematic joint angles are suppose to be slower movements much below 5Hz frequency
% Hence to avoid the natural trumer movements, we need to filter the
% signals. This code removes the trumers using 4th order zero lag low-pass
% filter with cutoff frequency as 5Hz. We using function BW_Filter for
% removing extra noise from the filter which lies above 5Hz. 

%% Code:
for Q=1:12
    addpath([pwd strcat('/Sub_',num2str(Q))]);
    X=zeros(1,96);
    for t=1:5 % Collecting all trials for single subject
        for i=1:70
            v=importdata(strcat('Sub_',num2str(Q),'_Posture',num2str(i),'_trial_',num2str(t),'_C.txt'));
            v_bore=mean(v(5:60,:));
            for a=1:length(v)
                Y(a,:)=v(a,:)-v_bore;
            end
            Y=Y(441:510,:); % selecting the data points from % 421:570
            X=[X;Y]; % 440-510: static (Joint angle), 420:571 (Static:input space), 60-540: Dynamic
        end
    end
    for t=1:5
        for i=1:30
            v=importdata(strcat('Sub_',num2str(Q),'_object',num2str(i),'_trial_',num2str(t),'_C.txt'));
            v_bore=mean(v(5:60,:));
            for a=1:length(v)
                Y(a,:)=v(a,:)-v_bore;
            end
            Y=Y(441:510,:); % selecting the data points from % 421:570
            X=[X;Y]; % 440-510: static, 60-540: Dynamic
        end
    end
    X=X(2:end,:);
    %% Seperate position and orientation for 'X' (single subject- all trials)
    % Orientation
    ori=zeros(length(X),1);
    for i=1:16
        ori=[ori,X(:,(i-1)*6+4),X(:,(i-1)*6+5),X(:,(i-1)*6+6)];
    end
    ori=ori(:,2:end);
    Index=ori(:,1:9); Index=(Index*pi)/180;
    Middle=ori(:,10:18); Middle=(Middle*pi)/180;
    Ring=ori(:,19:27); Ring=(Ring*pi)/180;
    Little=ori(:,28:36); Little=(Little*pi)/180;
    Thumb=ori(:,37:45); Thumb=(Thumb*pi)/180;
    WRT=ori(:,45:48); WRT=(WRT*pi)/180;
%     Derive joint angles
    % Index finger joint angles
    for i=1:length(Index)
        R1=angle2dcm(Index(i,1),Index(i,2),Index(i,3)); % computing direction cosine
        R2=angle2dcm(Index(i,4),Index(i,5),Index(i,6));
        R3=angle2dcm(Index(i,7),Index(i,8),Index(i,9));
        R=angle2dcm(WRT(i,1),WRT(i,2),WRT(i,3));
        R1=R1'; R2=R2'; R3=R3'; R=R'; % taking transformation (as given in polhemus)
        Rot_DIP=R1'*R2; Rot_PIP=R2'*R3; Rot_MCP=R3'*R;
        [D_the1(i,1),D_the2(i,1),D_the3(i,1)]=dcm2angle(Rot_DIP');
        [P_the1(i,1),P_the2(i,1),P_the3(i,1)]=dcm2angle(Rot_PIP');
        [M_the1(i,1),M_the2(i,1),M_the3(i,1)]=dcm2angle(Rot_MCP');
        clear R1; clear R2; clear R3; clear R; clear Rot_DIP; clear Rot_PIP; clear Rot_MCP;
    end
    I_DIP=[D_the1,D_the2,D_the3];
    I_PIP=[P_the1,P_the2,P_the3];
    I_MCP=[M_the1,M_the2,M_the3];
    clear D_the1; clear D_the2; clear D_the3;
    clear M_the1; clear M_the2; clear M_the3;
    clear P_the1; clear P_the2; clear P_the3;
    
    % Middle finger joint angles
    for i=1:length(Middle)
        R1=angle2dcm(Middle(i,1),Middle(i,2),Middle(i,3));
        R2=angle2dcm(Middle(i,4),Middle(i,5),Middle(i,6));
        R3=angle2dcm(Middle(i,7),Middle(i,8),Middle(i,9));
        R=angle2dcm(WRT(i,1),WRT(i,2),WRT(i,3));
        R1=R1'; R2=R2'; R3=R3'; R=R'; % taking transformation (as given in polhemus)
        Rot_DIP=R1'*R2; Rot_PIP=R2'*R3; Rot_MCP=R3'*R;
        [D_the1(i,1),D_the2(i,1),D_the3(i,1)]=dcm2angle(Rot_DIP');
        [P_the1(i,1),P_the2(i,1),P_the3(i,1)]=dcm2angle(Rot_PIP');
        [M_the1(i,1),M_the2(i,1),M_the3(i,1)]=dcm2angle(Rot_MCP');
        clear R1; clear R2; clear R3; clear R; clear Rot_DIP; clear Rot_PIP; clear Rot_MCP;
    end
    M_DIP=[D_the1,D_the2,D_the3];
    M_PIP=[P_the1,P_the2,P_the3];
    M_MCP=[M_the1,M_the2,M_the3];
    clear D_the1; clear D_the2; clear D_the3;
    clear M_the1; clear M_the2; clear M_the3;
    clear P_the1; clear P_the2; clear P_the3;
    
    % Ring finger joint angles
    for i=1:length(Ring)
        R1=angle2dcm(Ring(i,1),Ring(i,2),Ring(i,3));
        R2=angle2dcm(Ring(i,4),Ring(i,5),Ring(i,6));
        R3=angle2dcm(Ring(i,7),Ring(i,8),Ring(i,9));
        R=angle2dcm(WRT(i,1),WRT(i,2),WRT(i,3));
        R1=R1'; R2=R2'; R3=R3'; R=R'; % taking transformation (as given in polhemus)
        Rot_DIP=R1'*R2; Rot_PIP=R2'*R3; Rot_MCP=R3'*R;
        [D_the1(i,1),D_the2(i,1),D_the3(i,1)]=dcm2angle(Rot_DIP');
        [P_the1(i,1),P_the2(i,1),P_the3(i,1)]=dcm2angle(Rot_PIP');
        [M_the1(i,1),M_the2(i,1),M_the3(i,1)]=dcm2angle(Rot_MCP');
        clear R1; clear R2; clear R3; clear R; clear Rot_DIP; clear Rot_PIP; clear Rot_MCP;
    end
    R_DIP=[D_the1,D_the2,D_the3];
    R_PIP=[P_the1,P_the2,P_the3];
    R_MCP=[M_the1,M_the2,M_the3];
    clear D_the1; clear D_the2; clear D_the3;
    clear M_the1; clear M_the2; clear M_the3;
    clear P_the1; clear P_the2; clear P_the3;
    
    % Little finger joint angles
    for i=1:length(Little)
        R1=angle2dcm(Little(i,1),Little(i,2),Little(i,3));
        R2=angle2dcm(Little(i,4),Little(i,5),Little(i,6));
        R3=angle2dcm(Little(i,7),Little(i,8),Little(i,9));
        R=angle2dcm(WRT(i,1),WRT(i,2),WRT(i,3));
        R1=R1'; R2=R2'; R3=R3'; R=R'; % taking transformation (as given in polhemus)
        Rot_DIP=R1'*R2; Rot_PIP=R2'*R3; Rot_MCP=R3'*R;
        [D_the1(i,1),D_the2(i,1),D_the3(i,1)]=dcm2angle(Rot_DIP');
        [P_the1(i,1),P_the2(i,1),P_the3(i,1)]=dcm2angle(Rot_PIP');
        [M_the1(i,1),M_the2(i,1),M_the3(i,1)]=dcm2angle(Rot_MCP');
        clear R1; clear R2; clear R3; clear R; clear Rot_DIP; clear Rot_PIP; clear Rot_MCP;
    end
    L_DIP=[D_the1,D_the2,D_the3];
    L_PIP=[P_the1,P_the2,P_the3];
    L_MCP=[M_the1,M_the2,M_the3];
    clear D_the1; clear D_the2; clear D_the3;
    clear M_the1; clear M_the2; clear M_the3;
    clear P_the1; clear P_the2; clear P_the3;
    
    % Thumb joint angles
    for i=1:length(Thumb)
        R1=angle2dcm(Thumb(i,1),Thumb(i,2),Thumb(i,3));
        R2=angle2dcm(Thumb(i,4),Thumb(i,5),Thumb(i,6));
        R3=angle2dcm(Thumb(i,7),Thumb(i,8),Thumb(i,9));
        R=angle2dcm(WRT(i,1),WRT(i,2),WRT(i,3));
        R1=R1'; R2=R2'; R3=R3'; R=R'; % taking transformation (as given in polhemus)
        Rot_DIP=R1'*R2; Rot_PIP=R2'*R3; Rot_MCP=R3'*R;
        [D_the1(i,1),D_the2(i,1),D_the3(i,1)]=dcm2angle(Rot_DIP');
        [P_the1(i,1),P_the2(i,1),P_the3(i,1)]=dcm2angle(Rot_PIP');
        [M_the1(i,1),M_the2(i,1),M_the3(i,1)]=dcm2angle(Rot_MCP');
        clear R1; clear R2; clear R3; clear R; clear Rot_DIP; clear Rot_PIP; clear Rot_MCP;
    end
    TH_DIP=[D_the1,D_the2,D_the3];
    TH_PIP=[P_the1,P_the2,P_the3];
    TH_MCP=[M_the1,M_the2,M_the3];
    clear D_the1; clear D_the2; clear D_the3;
    clear M_the1; clear M_the2; clear M_the3;
    clear P_the1; clear P_the2; clear P_the3;
    
    J_angle=[I_DIP,I_PIP,I_MCP,M_DIP,M_PIP,M_MCP,R_DIP,R_PIP,R_MCP,L_DIP,L_PIP,L_MCP,TH_DIP,TH_PIP,TH_MCP];
    J_actual=(J_angle*180)/pi;
    clear Index, clear Middle; clear Ring; clear Litte; clear Thumb; clear WRT; clear J_angle;
%     select 21 out of 45
    I=[J_actual(:,1),J_actual(:,4),J_actual(:,7),J_actual(:,8)];
    M=[J_actual(:,10),J_actual(:,13),J_actual(:,16),J_actual(:,17)];
    R=[J_actual(:,19),J_actual(:,22),J_actual(:,25),J_actual(:,26)];
    L=[J_actual(:,28),J_actual(:,31),J_actual(:,34),J_actual(:,35)];
    Th=[J_actual(:,37),J_actual(:,40),J_actual(:,43),J_actual(:,44),J_actual(:,45)];
    X1=[I,M,R,L,Th]; % saved and reserved for comparing with predicted one
    
    %% Filtering the data at 4Hz frequency, Sampling frequency is 120Hz and cutoff frequency is 4Hz
    J_21_filter=BW_Filter(X1,120,4); % 4Hz cutoff for avoiding tremor frequency
    %% Save the filtered data below 4Hz
    save(strcat('Static_Jangle_filter4Hz_Sub_',num2str(Q),'.mat'),'J_21_filter');
    clear
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%  PART 2: Posture Similarity Index (PSI) %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Posture similarity index and eigen posture
clc;clear;
Q=input('Enter subject number: ');
load(strcat('Static_Jangle_filter4Hz_Sub_',num2str(Q),'.mat'));
J_21_actual=J_21_filter;
%% Constructing 100 postures dataset by selecting one among five trials
for S=1:5
    v=randi(5,1,70); % generate random sequence for selecting the posture trials, assuming no relation between trials.
    w=randi(5,1,30); % generate random sequence for selecting trials from 5 for 30 objects
    for i=1:70
        A{i}=J_21_actual((v(i)-1)*4900+(i-1)*70+1:(i*70)+((v(i)-1)*4900),:); % Change number of PCs based on amount of variance acounted for is 90% or greater
    end
    for j=1:30
        B{j}=J_21_actual((w(j)-1)*2100+(j-1)*70+1+24500:(j*70)+((w(j)-1)*2100+24500),:);
    end
    Q_pos=zeros(1,21);
    Q_obj=zeros(1,21);
    for i=1:70
        Q_pos=[Q_pos;A{1,i}];
    end
    for j=1:30
        Q_obj=[Q_obj;B{1,j}];
    end
    Q_pos=Q_pos(2:end,:);
    Q_obj=Q_obj(2:end,:);
    t{S}=[Q_pos;Q_obj];
    clear A; clear B; clear Q_pos; clear Q_obj;
end

%% Perform PCA of selected bag of data (7000x21 points)
x=randi(5,1,1); % select any random data bag

% Normaliation of data
Nome=bsxfun(@minus,t{x},min(t{x}));
range=max(t{x})-min(t{x});
J_norm=bsxfun(@rdivide,Nome,range);
[coef,score,latent,~,explained,mu]=pca(J_norm); % Project 21D data to 2d space using linear approach
Projected_x=score(:,1:21); % select projected data on PC space

%% Take the each posture seperately and find the projection of the PCs seperately and compute the covariance matrix for each posture seperately
for i=1:100
    COV_Q_PCA_data{i}=cov(Projected_x((i-1)*70+1:i*70,:));
end
exmpl_COV=COV_Q_PCA_data; % Copy the covariance matrix to other variable , if further processing is needed (Not used)

%% Posture similarity index (Compare the squared euclidean distance)
PSI_P=zeros(100);
for i=1:100
    for j=1:100
        for k=1:21
            Diffe=exmpl_COV{1,j}(k,k)-exmpl_COV{1,i}(k,k);
            squared_diff=Diffe*Diffe;
            PSI_P(i,j)=PSI_P(i,j)+latent(k,1)*squared_diff;
        end
    end
end
%% Finding the 'P' value, ratio of two index
% Min-Max normalization
for i=1:100
    for j=1:100
        PSI_1(i,j)=(PSI_P(i,j)-min(PSI_P(:,j)))/(max(PSI_P(:,j))-min(PSI_P(:,j)));
    end
end
% Max normailization (Varient)
PSI_temp=PSI_P;
% Elementation of diagonal elements
% for i=1:100
%     PSI_temp(i,i)=NaN;
% end
for i=1:100
    for j=1:100
        PSI_dissimilar(i,j)=PSI_temp(i,j)/max(PSI_temp(:,j)); % normalizing the dissimilar martix
    end
end
PSI_max_norm=1-PSI_dissimilar; % Convert dissimilar matrix into similar matrix;

%% Finding premitive postures (finding threshold for setting up the value for binary classification)
% Find the cumulative histogram
A=zeros(100,10);
for i=1:100 % individual posture wise
    for j=1:10
        A(i,j)=length(find(PSI_max_norm(:,i)<=j/10));
    end
end
B=sum(A); % Overall (across postures)
C=polyder(B); % taking first derivative of A, find the maximum value
[val,index_C]=max(C);
flag=index_C/10; % set threshold as peak of the derivative of cumulative histogram
disp(flag)
PSI_UP=PSI_max_norm;
PSI_UP(triu(PSI_UP==PSI_UP))=NaN;
% set the threshold values (Above threshold, value=1, below threshold value=0)
for i=1:100
    for j=1:100
        if PSI_UP(i,j)<flag
            PSI_UP(i,j)=0;
        else if PSI_UP(i,j)>=flag
                PSI_UP(i,j)=1;
            else
                PSI_UP(i,j)=NaN;
            end
        end
    end
end
% figure(2)
% pcolor(PSI_UP)
length(find(PSI_UP==1))
temp1=length(find(PSI_UP(1:11,1:11)==1)); % find no. of common postures
temp3=length(find(PSI_UP(71:100,71:100)==1)); % find no. of common postures
Special=length(find(PSI_UP(12:70,12:70)==1)); % find no. of special postures
Common=temp1+temp3;

%% Count the common postures and find which are similar postures having index>flag
for i=1:100
    W_count(i,1)=length(find(PSI_UP(:,i)==1));
    W_c{i}=find(PSI_UP(:,i)==1);
end

%% Finding Premitive postures
%% Approach 2: Columnwise posture importance
summ_PSI=sum(PSI_max_norm);
Total_PSI=sum(summ_PSI);
Relative_PII=summ_PSI/Total_PSI;
[PII_sorted(:,1),PII_sorted(:,2)]=sort(Relative_PII(1,:));
for i=1:100
    Cumu(i)=sum(PII_sorted(1:i,1));
end
% plot(polyder(Cumu),'r')
[e,ee]=max(polyder(Cumu))
Posture_TH=polyder(Cumu);
%% Representative postures 
RP2=PII_sorted(:,2); % PII_sorted contains index in the increaseing order based on information of all PSI
special=length(find(PII_sorted(ee:end,2)>11 & PII_sorted(ee:end,2)<71)); % find no. of special postures, have index between 12 and 70
common=100-(ee+special) % find no. of common postures that exist 
special

%% Saving the files/results, 
% PSI matrix (without half NaNs and with half NaNs), variance explained,
% Projected_x (on PCA space), Eigen vectors, mean value (mu)
%,no. of components higher than n (n from 0.1 upto 1) posturewise, Cummulative
% (B), derivative of cummulative (C), Threshold (index_C), PPI_sorted, Cumu
% (cumulative sum of all the PPI across postures), 

xlswrite(strcat('Linear synergy_sub-',num2str(Q),'.xls'),PSI_1,1); % without NaNs
xlswrite(strcat('Linear synergy_sub-',num2str(Q),'.xls'),PSI_UP,2); % with NaNs
xlswrite(strcat('Linear synergy_sub-',num2str(Q),'.xls'),explained,3); % variance
xlswrite(strcat('Linear synergy_sub-',num2str(Q),'.xls'),Projected_x,4); % projected on PC space
xlswrite(strcat('Linear synergy_sub-',num2str(Q),'.xls'),coef,5); % eigen vectors
xlswrite(strcat('Linear synergy_sub-',num2str(Q),'.xls'),mu,6); % Mean values for reconstruction
xlswrite(strcat('Linear synergy_sub-',num2str(Q),'.xls'),A,7); % histogram
xlswrite(strcat('Linear synergy_sub-',num2str(Q),'.xls'),B,8); % cummulative hist
xlswrite(strcat('Linear synergy_sub-',num2str(Q),'.xls'),C,9); % derivative of cummulative hist (Plot)
xlswrite(strcat('Linear synergy_sub-',num2str(Q),'.xls'),index_C,10); %Threshold
% xlswrite(strcat('Linear synergy_sub-',num2str(Q),'.xls'),Special,11); %no. of special postures shows similarity greater than index_C
% xlswrite(strcat('Linear synergy_sub-',num2str(Q),'.xls'),Common,12); %Common postures count that shows similarity
xlswrite(strcat('Linear synergy_sub-',num2str(Q),'.xls'),PII_sorted,11); %contains sorted Postures in ascending order
xlswrite(strcat('Linear synergy_sub-',num2str(Q),'.xls'),Cumu,12); %cumulative summed values
xlswrite(strcat('Linear synergy_sub-',num2str(Q),'.xls'),ee,13,'A1'); %NO of postures that can be explained
xlswrite(strcat('Linear synergy_sub-',num2str(Q),'.xls'),common,13,'B1'); %no of common postures
xlswrite(strcat('Linear synergy_sub-',num2str(Q),'.xls'),special,13,'C1'); %no of special postures
xlswrite(strcat('Linear synergy_sub-',num2str(Q),'.xls'),Posture_TH,14); %derivative of posture threshold (Plot)